import { F } from "../api/types.js";
import { builtinCall } from "../ast/function.js";
import { matchingBoolType, matchingPrimFor } from "../ast/item.js";
const primOp1 = (name) => (a) => builtinCall(name, a.type, a);
const primOp2 = (name) => (a, b) => builtinCall(name, a.type, a, b);
const primOp3 = (name) => (a, b, c) => builtinCall(name, a.type, a, b, c);
const normalize = (v) => builtinCall("normalize", v.type, v);
const length = (v) => builtinCall("length", F, v);
const distance = (a, b) => builtinCall("distance", F, a, b);
const dot = (a, b) => builtinCall("dot", F, a, b);
const cross = (a, b) => builtinCall("cross", a.type, a, b);
const reflect = (i, n) => builtinCall("reflect", i.type, i, n);
const refract = (i, n, ior) => builtinCall("refract", i.type, i, n, ior);
const faceForward = (i, n, nref) => builtinCall("faceForward", i.type, i, n, nref);
const min = primOp2("min");
const max = primOp2("max");
const clamp = primOp3("clamp");
const step = primOp2("step");
const smoothstep = primOp3("smoothstep");
const radians = primOp1("radians");
const degrees = primOp1("degrees");
const cos = primOp1("cos");
const sin = primOp1("sin");
const tan = primOp1("tan");
const acos = primOp1("acos");
const asin = primOp1("asin");
function atan(a, b) {
  const f = b ? builtinCall("atan", a.type, a, b) : builtinCall("atan", a.type, a);
  b && (f.info = "nn");
  return f;
}
const pow = primOp2("pow");
const exp = primOp1("exp");
const log = primOp1("log");
const exp2 = primOp1("exp2");
const log2 = primOp1("log2");
const sqrt = primOp1("sqrt");
const inversesqrt = primOp1("inversesqrt");
const abs = primOp1("abs");
const sign = primOp1("sign");
const floor = primOp1("floor");
const ceil = primOp1("ceil");
const fract = primOp1("fract");
const powf = (x, y) => pow(x, matchingPrimFor(x, y));
function mod(a, b) {
  const f = builtinCall("mod", a.type, a, b);
  b.type === F && (f.info = "n");
  return f;
}
const modf = (a, b) => builtinCall("modf", a.type, a, b);
function mix(a, b, c) {
  const f = builtinCall("mix", a.type, a, b, c);
  c.type === F && (f.info = "n");
  return f;
}
const matrixCompMult = (a, b) => builtinCall("matrixCompMult", a.type, a, b);
function isnan(a) {
  return builtinCall("isnan", matchingBoolType(a), a);
}
function isinf(a) {
  return builtinCall("isinf", matchingBoolType(a), a);
}
export {
  abs,
  acos,
  asin,
  atan,
  ceil,
  clamp,
  cos,
  cross,
  degrees,
  distance,
  dot,
  exp,
  exp2,
  faceForward,
  floor,
  fract,
  inversesqrt,
  isinf,
  isnan,
  length,
  log,
  log2,
  matrixCompMult,
  max,
  min,
  mix,
  mod,
  modf,
  normalize,
  pow,
  powf,
  radians,
  reflect,
  refract,
  sign,
  sin,
  smoothstep,
  sqrt,
  step,
  tan
};
